
/* packet-facn.c
 * Routines for FACN packet disassembly
 *
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <string.h>
#include <glib.h>
#include <epan/packet.h>
#include <epan/strutil.h>
#include <epan/resolv.h>
#include "packet-arp.h"
#include "etypes.h"



static int proto_facn = -1;
static int hf_facn_tcd = -1;
static int hf_facn_ml = -1;
static int hf_facn_sna = -1;
static int hf_facn_dna = -1;
static int hf_facn_vseq = -1;
static int hf_facn_seq = -1;
static int hf_facn_mctl = -1;
static int hf_facn_uls = -1;
static int hf_facn_mwsz = -1;
static int hf_facn_madd = -1;
static int hf_facn_mft = -1;
static int hf_facn_mrlt = -1;
static int hf_facn_ver = -1;
static int hf_facn_cad1 = -1;
static int hf_facn_csz1 = -1;
static int hf_facn_cad2 = -1;
static int hf_facn_csz2 = -1;
static int hf_facn_mode = -1;
static int hf_facn_pver = -1;
static int hf_facn_pri = -1;
static int hf_facn_cbn = -1;
static int hf_facn_tbn = -1;
static int hf_facn_bsize = -1;
static int hf_facn_lks = -1;
static int hf_facn_tw = -1;
static int hf_facn_rct = -1;
static int hf_facn_ndn = -1;
static int hf_facn_vdn = -1;
static int hf_facn_msn = -1;

/* uls bit */
static int hf_facn_ulsrun = -1;
static int hf_facn_ulsstop = -1;
static int hf_facn_ulsEcode = -1;
static int hf_facn_ulsUerr = -1;
/* LKS bit */
static int hf_facn_lksb04 = -1;
static int hf_facn_lksb05 = -1;
static int hf_facn_lksb06 = -1;
static int hf_facn_lksb07 = -1;


static gint ett_facn = -1;
static gint ett_facn_uls = -1;
static gint ett_facn_lks = -1;

/*TCD Transaction Code */
#define TCD_Transparent                       59999
#define TCD_TokenHeader                       65000
#define TCD_CyclicHeader                      65001
#define TCD_EntryRequestFrameHeader           65002
#define TCD_RequestOfReadDataFromByteBlock    65003
#define TCD_RequestOfWriteDataFromByteBlock   65004
#define TCD_RequestOfReadDataFromWordBlock    65005
#define TCD_RequestOfWriteDataFromWordBlock   65006
#define TCD_RequestOfReadParameterOfNetwork   65007
#define TCD_RequestOfWriteParameterOfNetwork  65008
#define TCD_RequestOfStopCommand              65009
#define TCD_RequestOfStartCommand             65010
#define TCD_RequestOfReadOfProfile            65011
#define TCD_EntryTriggerFrameHeader           65012
#define TCD_RequestOfLogFrame                 65013
#define TCD_RequsetOfLogClear                 65014

/*TCD(Message Response)*/
#define TCD_ResponseOfReadDataFromByteBlock   65203
#define TCD_ResponseOfWriteDataFromByteBlock  65204
#define TCD_ResponseOfReadDataFromWordBlock   65205
#define TCD_ResponseOfWriteDataFromWordBlock  65206
#define TCD_ResponseOfReadParameterOfNetwork  65207
#define TCD_ResponseOfWriteParameterOfNetwork 65208
#define TCD_ResponseOfStopCommand             65209
#define TCD_ResponseOfStartCommand            65210
#define TCD_ResponseOfReadOfProfile           65211
#define TCD_ResponseOfLogFrame                65213
#define TCD_ResponseOfLogClear                65214

/* Param Offset (form FACN Header) */
#define ML_OFFSET    4
#define SA_OFFSET    8
#define SNA_OFFSET  SA_OFFSET+3
#define DA_OFFSET   12
#define DNA_OFFSET  DA_OFFSET+3
#define VSEQ_OFFSET 16
#define SEQ_OFFSET  20
#define MCTL_OFFSET 24
#define ULS_OFFSET  28
#define MWSZ_OFFSET 30
#define MADD_OFFSET 32
#define MFT_OFFSET  36
#define MRLT_OFFSET 37
#define TCD_OFFSET  40
#define VER_OFFSET  42
#define CAD1_OFFSET 44
#define CSZ1_OFFSET 46
#define CAD2_OFFSET 48
#define CSZ2_OFFSET 50
#define MODE_OFFSET 52
#define PVER_OFFSET 54
#define PRI_OFFSET  55
#define CBN_OFFSET  56
#define TBN_OFFSET  57
#define BSIZE_OFFSET 58
#define LKS_OFFSET  60
#define TW_OFFSET   61
#define RCT_OFFSET  62

#define NDN_OFFSET 64
#define VDN_OFFSET 74
#define MSN_OFFSET 84

/*MASK*/
#define ULS_RUN_MASK      0x8000
#define ULS_UERR_MASK     0x6000
#define ULS_ERR_CODE_MASK 0x0FFF

#define LKS_UPPER_ERROR_MASK 0x10
#define LKS_COM_VALID_MASK   0x20
#define LKS_COM_CMPT_MASK    0x40
#define LKS_COM_DUP_MASK     0x80

/* */
#define NDN_LENGTH 10
#define VDN_LENGTH 10
#define MSN_LENGTH 10

/*
 *   Get TCD String.
 */
static void
facn_get_TcdName(guint tcd,gchar *strBuff)
{
	if((0 <= tcd) && (tcd <= TCD_Transparent)){
		strcpy(strBuff, "Transparent");
		return;
	}
	
	switch(tcd){
		case TCD_Transparent:
			strcpy(strBuff, "Transparent");
			break;
		case TCD_TokenHeader:
			strcpy(strBuff, "TokenHeader");
			break;
		case TCD_CyclicHeader:
			strcpy(strBuff, "CyclicHeader");
			break;
		case TCD_EntryRequestFrameHeader:
			strcpy(strBuff, "EntryRequestFrameHeader");
			break;
		case TCD_RequestOfReadDataFromByteBlock:
			strcpy(strBuff, "RequestOfReadDataFromByteBlock");
			break;
		case TCD_RequestOfWriteDataFromByteBlock:
			strcpy(strBuff, "RequestOfWriteDataFromByteBlock");
			break;
		case TCD_RequestOfReadDataFromWordBlock:
			strcpy(strBuff, "RequestOfReadDataFromWordBlock");
			break;
		case TCD_RequestOfWriteDataFromWordBlock:
			strcpy(strBuff, "RequestOfWriteDataFromWordBlock");
			break;
		case TCD_RequestOfReadParameterOfNetwork:
			strcpy(strBuff, "RequestOfReadParameterOfNetwork");
			break;
		case TCD_RequestOfWriteParameterOfNetwork:
			strcpy(strBuff, "RequestOfWriteParameterOfNetwork");
			break;
		case TCD_RequestOfStopCommand:
			strcpy(strBuff, "RequestOfStopCommand");
			break;
		case TCD_RequestOfStartCommand:
			strcpy(strBuff, "RequestOfStartCommand");
			break;
		case TCD_RequestOfReadOfProfile:
			strcpy(strBuff, "RequestOfReadOfProfile");
			break;
		case TCD_EntryTriggerFrameHeader:
			strcpy(strBuff, "EntryTriggerFrameHeader");
			break;
		case TCD_RequestOfLogFrame:
			strcpy(strBuff, "RequestOfLogFrame");
			break;
		case TCD_RequsetOfLogClear:
			strcpy(strBuff, "RequsetOfLogClear");
			break;

		/*TCD(Message Response)*/
		case TCD_ResponseOfReadDataFromByteBlock:
			strcpy(strBuff, "ResponseOfReadDataFromByteBlock");
			break;
		case TCD_ResponseOfWriteDataFromByteBlock:
			strcpy(strBuff, "ResponseOfWriteDataFromByteBlock");
			break;
		case TCD_ResponseOfReadDataFromWordBlock:
			strcpy(strBuff, "ResponseOfReadDataFromWordBlock");
			break;
		case TCD_ResponseOfWriteDataFromWordBlock:
			strcpy(strBuff, "ResponseOfWriteDataFromWordBlock");
			break;
		case TCD_ResponseOfReadParameterOfNetwork:
			strcpy(strBuff, "ResponseOfReadParameterOfNetwork");
			break;
		case TCD_ResponseOfWriteParameterOfNetwork:
			strcpy(strBuff, "ResponseOfWriteParameterOfNetwork");
			break;
		case TCD_ResponseOfStopCommand:
			strcpy(strBuff, "ResponseOfStopCommand");
			break;
		case TCD_ResponseOfStartCommand:
			strcpy(strBuff, "ResponseOfStartCommand");
			break;
		case TCD_ResponseOfReadOfProfile:
			strcpy(strBuff, "ResponseOfReadOfProfile");
			break;
		case TCD_ResponseOfLogFrame:
			strcpy(strBuff, "ResponseOfLogFrame");
			break;
		case TCD_ResponseOfLogClear:
			strcpy(strBuff, "ResponseOfLogClear");
			break;
		default:
			strcpy(strBuff, "Undefined Frame");
			break;
	}
}
static void
dissect_facn_entry_frame(tvbuff_t *tvb, packet_info *pinfo, proto_tree *facn_tree)
{
	guint8 name[16];
	
	/* NDN */
	memset( name, '\0', sizeof(name));
	tvb_get_nstringz0(tvb, NDN_OFFSET, NDN_LENGTH, name);
	proto_tree_add_uint_format(facn_tree, hf_facn_ndn, tvb, NDN_OFFSET, NDN_LENGTH, 
					name[0],
					"NDN: %s",
					name);
					
	/* VDN */
	memset( name, '\0', sizeof(name));
	tvb_get_nstringz0(tvb, VDN_OFFSET, VDN_LENGTH, name);
	proto_tree_add_uint_format(facn_tree, hf_facn_vdn, tvb, VDN_OFFSET, VDN_LENGTH, 
					name[0],
					"VDN: %s",
					name);
					
	/* MDN */
	memset( name, '\0', sizeof(name));
	tvb_get_nstringz0(tvb, MSN_OFFSET, MSN_LENGTH, name);
	proto_tree_add_uint_format(facn_tree, hf_facn_msn, tvb, MSN_OFFSET, MSN_LENGTH, 
					name[0],
					"MSN: %s",
					name);
	
}
static void
dissect_facn(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	proto_tree *facn_tree,*field_tree;
	proto_item *ti,*tf;
	guint tcd;
	guint32 ml,vseq,seq,mctl,madd;
	guint16 uls,mwsz,ver,c_addr1,c_addr2,c_size1,c_size2,mode;
	guint16 mft,mrlt,bsize,rct;
	guint8  sna,dna,pver,pri,cbn,tbn,lks,tw;
	
	gchar      type_str[64],col_info[64],nodeNum[8];
	guint      length, reported_length;
	
	/* set Protocol Type */
	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "FACN");
		
	/* get TCD Name */
	tcd = tvb_get_ntohs(tvb, TCD_OFFSET);
	facn_get_TcdName(tcd,type_str);
	
	/* Set Protocol Infomation*/
	sna = tvb_get_guint8(tvb, SNA_OFFSET);
	dna = tvb_get_guint8(tvb, DNA_OFFSET);
	strcpy(col_info,type_str);
	strcat(col_info,"(");
	sprintf(nodeNum,"%u",sna);
	strcat(col_info,nodeNum);
	strcat(col_info,"->");
	sprintf(nodeNum,"%u",dna);
	strcat(col_info,nodeNum);
	strcat(col_info,")");
	if (check_col(pinfo->cinfo, COL_INFO))
		col_add_str(pinfo->cinfo, COL_INFO, col_info);
	
	/* set Tree View */
	if (tree) {
		length = tvb_length(tvb);
		reported_length = tvb_reported_length(tvb);
		ti = proto_tree_add_item(tree, proto_facn, tvb, 0, length, FALSE);
		facn_tree = proto_item_add_subtree(ti, ett_facn);
		
		/* TCD */
		proto_tree_add_uint_format(facn_tree, hf_facn_tcd, tvb, TCD_OFFSET, 2, 
			       tcd,
			       "TCD: %u (%s)",
			       tcd, type_str);
		/* ML */
		ml = tvb_get_ntohl(tvb, ML_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_ml, tvb, ML_OFFSET, 4, 
			       ml,
			       "ML: %u",
			       ml);
		/* SNA */
		proto_tree_add_uint_format(facn_tree, hf_facn_sna, tvb, SNA_OFFSET, 1, 
			       sna,
			       "SNA(Src Node Address): %u ",
			       sna);
		/* DNA */
		proto_tree_add_uint_format(facn_tree, hf_facn_dna, tvb, DNA_OFFSET + 3, 1, 
			       dna,
			       "DNA(Dst Node Address): %u ",
			       dna);
		/* VSEQ */
		vseq = tvb_get_ntohl(tvb, VSEQ_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_vseq, tvb, VSEQ_OFFSET, 4, 
			       vseq,
			       "VSEQ: %u",
			       vseq);
		/* SEQ */
		seq = tvb_get_ntohl(tvb, SEQ_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_seq, tvb, SEQ_OFFSET, 4, 
			       seq,
			       "SEQ: %u",
			       seq);
		/* MCTL */
		mctl = tvb_get_ntohl(tvb, MCTL_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_mctl, tvb, MCTL_OFFSET, 4, 
			       mctl,
			       "MCTL: %u",
			       mctl);
		/*
		 * ULS
		 */
		uls = tvb_get_ntohs(tvb, ULS_OFFSET);
		tf = proto_tree_add_uint_format(facn_tree, hf_facn_uls, tvb, ULS_OFFSET, 2, 
			       uls,
			       "ULS(Upper Layer Status): %u",
			       uls);
		field_tree = proto_item_add_subtree(tf, ett_facn_uls);
		
		if((uls & ULS_UERR_MASK) == 0x0000 ){
			/* NOLMAL */
			strcpy(type_str, "NOLMAL");
		} else if((uls & ULS_UERR_MASK) == 0x2000){
			/* WARNING */
			strcpy(type_str, "WARNING");
		} else {
			/* ALARM */
			strcpy(type_str, "ALARM");
		}
		proto_tree_add_uint_format(field_tree, hf_facn_ulsUerr, tvb, ULS_OFFSET, 4, 
			       uls,
			       "U_ERR: %s",
			       type_str);
		
		proto_tree_add_uint(field_tree, hf_facn_ulsEcode, tvb, ULS_OFFSET, 2, uls);
		
		if(uls & ULS_RUN_MASK){
			/* RUN */
			proto_tree_add_uint(field_tree, hf_facn_ulsrun, tvb, ULS_OFFSET, 2, uls);
		} else {
			/* STOP */
			proto_tree_add_uint(field_tree, hf_facn_ulsstop, tvb, ULS_OFFSET, 2, uls);
		}

		/* MWSZ */
		mwsz = tvb_get_ntohs(tvb, MWSZ_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_mwsz, tvb, MWSZ_OFFSET, 2, 
			       mwsz,
			       "MWSZ: %u",
			       mwsz);
		/* MADD */
		madd = tvb_get_ntohl(tvb, MADD_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_madd, tvb, MADD_OFFSET, 4, 
			       madd,
			       "MADD: %u",
			       madd);
		/* MFT */
		mft = tvb_get_guint8(tvb, MFT_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_mft, tvb, MFT_OFFSET, 1, 
			       mft,
			       "MFT: %u msec",
			       mft);
		/*
		 * M_RLT
		 */
		mrlt = tvb_get_guint8(tvb, MRLT_OFFSET);
		if(mrlt == 0x00 ){
			/* NOLMAL */
			strcpy(type_str, "Nomal Response");
		} else if(mrlt == 0x01){
			/* Error */
			strcpy(type_str, "Error Response");
		} else if(mrlt == 0x02){
			/* No Support */
			strcpy(type_str, "No Support Response");
		} else {
			strcpy(type_str, "Undefine Code ?");
		}
		proto_tree_add_uint_format(facn_tree, hf_facn_mrlt, tvb, MRLT_OFFSET, 1, 
			       mrlt,
			       "MRLT: %s (%u)",
			       type_str,mrlt);
		/* VER */
		ver = tvb_get_ntohs(tvb, VER_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_ver, tvb, VER_OFFSET, 2, 
			       ver,
			       "VER: %u",
			       ver);
		/* C_AD1 */
		c_addr1 = tvb_get_ntohs(tvb, CAD1_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_cad1, tvb, CAD1_OFFSET, 2, 
			       c_addr1,
			       "Common Memory1 Address: %u",
			       c_addr1);
		/* C_SZ1 */
		c_size1 = tvb_get_ntohs(tvb, CSZ1_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_csz1, tvb, CSZ1_OFFSET, 2, 
			       c_size1,
			       "Common Memory1 Size: %u",
			       c_size1);
		/* C_AD2 */
		c_addr2 = tvb_get_ntohs(tvb, CAD2_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_cad2, tvb, CAD2_OFFSET, 2, 
			       c_addr2,
			       "Common Memory2 Address: %u",
			       c_addr2);
		/* C_SZ2 */
		c_size2 = tvb_get_ntohs(tvb, CSZ2_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_csz2, tvb, CSZ2_OFFSET, 2, 
			       c_size2,
			       "Common Memory2 Size: %u",
			       c_size2);
		/* MODE */
		mode = tvb_get_ntohs(tvb, MODE_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_mode, tvb, MODE_OFFSET, 2, 
			       mode,
			       "MODE: %u",
			       mode);
		/* PVER */
		pver = tvb_get_guint8(tvb, PVER_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_pver, tvb, PVER_OFFSET, 1, 
			       pver,
			       "PVER: %u",
			       pver);
		/* PRI */
		pri = tvb_get_guint8(tvb, PRI_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_pri, tvb, PRI_OFFSET, 1, 
			       pri,
			       "PRI: %u",
			       pri);
		/* CBN */
		cbn = tvb_get_guint8(tvb, CBN_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_cbn, tvb, CBN_OFFSET, 1, 
			       cbn,
			       "CBN: %u",
			       cbn);
		/* TBN */
		tbn = tvb_get_guint8(tvb, TBN_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_tbn, tvb, TBN_OFFSET, 1, 
			       tbn,
			       "TBN: %u",
			       tbn);
		/* BSIZE */
		bsize = tvb_get_ntohs(tvb, BSIZE_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_bsize, tvb, BSIZE_OFFSET, 2, 
			       bsize,
			       "BSIZE: %u",
			       bsize);
		/*
		 * LKS
		 */
		lks = tvb_get_guint8(tvb, LKS_OFFSET);
		tf = proto_tree_add_uint_format(facn_tree, hf_facn_lks, tvb, LKS_OFFSET, 1, 
			       lks,
			       "LKS(Link Status): %u",
			       lks);
		field_tree = proto_item_add_subtree(tf, ett_facn_lks);
		proto_tree_add_uint(field_tree, hf_facn_lksb07, tvb, LKS_OFFSET, 1, lks);
		proto_tree_add_uint(field_tree, hf_facn_lksb06, tvb, LKS_OFFSET, 1, lks);
		proto_tree_add_uint(field_tree, hf_facn_lksb05, tvb, LKS_OFFSET, 1, lks);
		proto_tree_add_uint(field_tree, hf_facn_lksb04, tvb, LKS_OFFSET, 1, lks);
		
		/* TW */
		tw = tvb_get_guint8(tvb, TW_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_tw, tvb, TW_OFFSET, 1, 
			       tw,
			       "TW: %u msec",
			       tw);
		/* RCT */
		rct = tvb_get_ntohs(tvb, RCT_OFFSET);
		proto_tree_add_uint_format(facn_tree, hf_facn_rct, tvb, RCT_OFFSET, 2, 
			       rct,
			       "RCT: %u",
			       rct);
		/*
		 *  If this frame is Entry or Trigger, add some trees.
		 */
		if( (tcd == TCD_EntryRequestFrameHeader) || (tcd == TCD_EntryTriggerFrameHeader) ){
			dissect_facn_entry_frame(tvb, pinfo, facn_tree);
		}
	}
}
void
proto_register_facn(void)
{
	static hf_register_info hf[] = {
		{ &hf_facn_tcd,
			{ "Tcd", "facn.tcd", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_ml,
			{ "ML", "facn.ml", FT_UINT32, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_sna,
			{ "SNA", "facn.sna", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_dna,
			{ "DNA", "facn.dna", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_vseq,
			{ "VSEQ", "facn.vseq", FT_UINT32, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_seq,
			{ "SEQ", "facn.seq", FT_UINT32, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_mctl,
			{ "MCTL", "facn.mctl", FT_UINT32, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		/* ULS */
		{ &hf_facn_uls,
			{ "ULS", "facn.uls", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_ulsEcode,
			{ "User Error Code ", "facn.ulsEcode", FT_UINT16, BASE_DEC,	NULL, ULS_ERR_CODE_MASK, "", HFILL }
		},
		{ &hf_facn_ulsUerr,
			{ "U_ERR", "facn.ulsUerr", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_ulsrun,
			{ "RUN", "facn.ulsrun", FT_UINT16, BASE_DEC,	NULL, ULS_RUN_MASK, "", HFILL }
		},
		{ &hf_facn_ulsstop,
			{ "STOP", "facn.ulsstop", FT_UINT16, BASE_DEC,	NULL, ULS_RUN_MASK, "", HFILL }
		},
		
		{ &hf_facn_mwsz,
			{ "MWSZ", "facn.mwsz", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_madd,
			{ "MADD", "facn.madd", FT_UINT32, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_mft,
			{ "MFT", "facn.mft", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_mrlt,
			{ "MRLT", "facn.mrlt", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_ver,
			{ "VER", "facn.ver", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_cad1,
			{ "C_AD1", "facn.cad1", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_csz1,
			{ "C_SZ1", "facn.csz1", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_cad2,
			{ "C_AD2", "facn.cad2", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_csz2,
			{ "C_SZ2", "facn.csz2", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_mode,
			{ "MODE", "facn.mode", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_pver,
			{ "PVER", "facn.pver", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_pri,
			{ "PRI", "facn.pri", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_cbn,
			{ "CBN", "facn.cbn", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_tbn,
			{ "TBN", "facn.tbn", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_bsize,
			{ "BSIZE", "facn.bsize", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		/*
		 * LKS
		 */
		{ &hf_facn_lks,
			{ "LKS", "facn.lks", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_lksb04,
			{ "Upper Layer Error Signal", "facn.lksULErr", FT_UINT8, BASE_DEC,NULL, LKS_UPPER_ERROR_MASK, "", HFILL }
		},
		{ &hf_facn_lksb05,
			{ "Common Memory Data Valid", "facn.lksComValid", FT_UINT8, BASE_DEC,	NULL, LKS_COM_VALID_MASK, "", HFILL }
		},
		{ &hf_facn_lksb06,
			{ "Common Memory Init Complete", "facn.lksComInt", FT_UINT8, BASE_DEC,	NULL, LKS_COM_CMPT_MASK, "", HFILL }
		},
		{ &hf_facn_lksb07,
			{ "Address Duplication", "facn.lksAddDup", FT_UINT8, BASE_DEC,	NULL, LKS_COM_DUP_MASK, "", HFILL }
		},
		
		
		{ &hf_facn_tw,
			{ "TW", "facn.tw", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_rct,
			{ "RCT", "facn.rct", FT_UINT16, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_ndn,
			{ "NDN", "facn.ndn", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_vdn,
			{ "VDN", "facn.vdn", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		},
		{ &hf_facn_msn,
			{ "MDN", "facn.mdn", FT_UINT8, BASE_DEC,	NULL, 0x0, "", HFILL }
		}
	};
	
	static gint *ett[] = {
		&ett_facn,
		&ett_facn_uls,
		&ett_facn_lks
	};
	
	proto_facn = proto_register_protocol("FA Link Protocol", "FACN", "facn");
	proto_register_field_array(proto_facn, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
  
}

void
proto_reg_handoff_facn(void)
{
	dissector_handle_t facn_handle;

	/*
	 * Get handle for the UDP dissector.
	 */

	facn_handle = create_dissector_handle(dissect_facn, proto_facn);
/*	dissector_add("udp.port", 55001, facn_handle);*/
	dissector_add("udp.port", 55003, facn_handle);

}

